# automatically generated by the FlatBuffers compiler, do not modify

# namespace: proto

# WAMP transport channel type (`channel_type`). WAMP can run over any Transport which is message-based (requiring no WAMP transport channel framing) or stream-based (requiring WAMP transport channel framing), and bidirectional, reliable and ordered.
class TransportChannelType(object):
    # Not set or applicable.
    NULL = 0
    # Host language native function call transport, e.g. inherently message-based already (function call).
    FUNCTION = 1
    # Host run-time / OS level in-memory transport, e.g. (unframed / stream-based) memory buffer.
    MEMORY = 2
    # Serial (UART) based transport (unframed / stream-based).
    SERIAL = 3
    # TCP (non-TLS) based transport (unframed / stream-based).
    TCP = 4
    # TLS (over TCP) based transport (unframed / stream-based).
    TLS = 5
