# automatically generated by the FlatBuffers compiler, do not modify

# namespace: proto

# WAMP message type IDs organized by message categories:
# - Category 1: Neither Payload nor Forwarding (12 messages)
# - Category 2: Payload Only (0 messages) - architecturally empty
# - Category 3: Forwarding Only (3 messages)
# - Category 4: Both Payload and Forwarding (7 messages)
class MessageType(object):
    # No valid message type
    NULL = 0
    # [Category 1] Session opening, authentication and closing
    HELLO = 1
    WELCOME = 2
    ABORT = 3
    CHALLENGE = 4
    AUTHENTICATE = 5
    GOODBYE = 6
    # [Category 4] Used in both PubSub and RPC
    ERROR = 8
    # Publish & Subscribe (PubSub)
    PUBLISH = 16
    PUBLISHED = 17
    SUBSCRIBE = 32
    SUBSCRIBED = 33
    UNSUBSCRIBE = 34
    UNSUBSCRIBED = 35
    EVENT = 36
    EVENT_RECEIVED = 37
    # Remote Procedure Calls (RPC)
    CALL = 48
    CANCEL = 49
    RESULT = 50
    REGISTER = 64
    REGISTERED = 65
    UNREGISTER = 66
    UNREGISTERED = 67
    INVOCATION = 68
    INTERRUPT = 69
    YIELD = 70
