# automatically generated by the FlatBuffers compiler, do not modify

# namespace: proto

# WAMP authentication method (see: WAMP protocol spec, "13. Authentication Methods").
class AuthMethod(object):
    # Not set / not authenticated: `anonymous`.
    NULL = 0
    # WAMP Ticket authentication: `ticket` (see: WAMP protocol spec, "13.1. Ticket-based Authentication").
    TICKET = 1
    # WAMP Challenge-Response authentication: `wampcra` (see: WAMP protocol spec, "13.2. Challenge Response Authentication").
    CRA = 2
    # WAMP Salted Challenge Response authentication: `wamp-scram` (see: WAMP protocol spec, "13.3. Salted Challenge Response Authentication").
    SCRAM = 3
    # WAMP Cryptosign authentication: `cryptosign` (see: WAMP protocol spec, "13.4. Cryptosign-based Authentication").
    CRYPTOSIGN = 4
