# automatically generated by the FlatBuffers compiler, do not modify

# namespace: proto

# WAMP Payload Passthru (PPT) application payload serializer (renamed from `Serializer`).
class PPTSerializer(object):
    # Use same serializer (dynamically or statically typed) for the application payload as used on the transport for the WAMP message envelope.
    TRANSPORT = 0
    # Use JSON serializer for dynamically typed application payload.
    JSON = 1
    # Use MsgPack serializer for dynamically typed application payload.
    MSGPACK = 2
    # Use CBOR serializer for dynamically typed application payload.
    CBOR = 3
    # Use UBJSON serializer for dynamically typed application payload.
    UBJSON = 4
    # Raw pass-through of application payload, uninterpreted in any way.
    OPAQUE = 5
    # Use FlatBuffers serializer for statically typed application payload (https://google.github.io/flatbuffers/index.html).
    FLATBUFFERS = 6
    # Use FlexBuffers serializer dynamically typed application payload (https://google.github.io/flatbuffers/flexbuffers.html).
    FLEXBUFFERS = 7
